/*
 *  xnflib.h
 *
 *  Created by Francois Delyon on 03.03.07.
 *  Copyright 2007. All rights reserved.
 *  Version 2.0
 *
 */
 
 /*
	xnflib provides access to xnf format (documentation about the xnf format can be found at "http://www.satimage.fr/software/en/sl_xnfv2.html").
	An xnf file is a directory (a bundle in macosx) containing a table of contents stored as an xml file (index.xml).
	The table of contents (or "index") describes the contents of the binary data files stored in the subdirectory 'Contents'.
	This implementation allows to read and write arrays of number and the associated scales.
	This implementation takes into account the byte order of the platform.
	
	The main goal of the xnf format is to provide a standard access to binary files used in numerical computations.
	By design, it is straightforward to register existing binary files into the index either with standard xml tools or
	with xnflib (xxxByRef functions) or with a simple text editor.
	
	This library relies on the open source project libxml2: "http://xmlsoft.org/". Version >=2.6.16.
	On macosx, libxml2 is pre-installed.
 */
 
#ifndef __XNFLIB__
#define __XNFLIB__
#ifndef SATIMAGE
#include <libxml/tree.h>
#include <libxml/uri.h>
#else
#include <tree.h>
#include <uri.h>
#endif
#ifdef __cplusplus
extern "C" {
#endif

/*MaxDim is the maximal number of dimensions of arrays. 
MaxDim=5 should be suitable in most cases, but this value can be modified at your convenience. */
#define MaxDim 5

enum NumberType{
	TypeDouble=0,
	TypeFloat,
	TypeInteger,
	TypeShort,
	TypeChar,	
	TypeUInteger,
	TypeUShort,
	TypeUChar,	
	TypeDoubleComplex,	
	TypeFloatComplex	
};
/*xnfTypeSize[theNumberType] returns the size of elements of NumberType theNumberType */
extern int xnfTypeSize[10];
/*currentByteOrder 0: little endian, 1: big endian*/
extern int currentByteOrder;

/* xnfCreate
Create a new xnf file
@path: path to the new file
@ref: output -> a reference to the xml document
Result: 0 on success
*/
int xnfCreate(const char*path, xmlDocPtr* ref);

/* xnfOpen
Open an existing xnf file
@path: path to the existing file
@ref: output -> a reference to the xml document
Result: 0 on success
*/
int xnfOpen(const char*path, xmlDocPtr* ref);

/* xnfSave
Save the index of xnf file
@ref: the xml document (provided by xnfOpen or xnfCreate)
Result: 0 on success
*/
int xnfSave(xmlDocPtr ref);


/* xnfClose
Release the memory. After xnfClose ref becomes invalid.
Don't forget to call xnfSave(ref) (before a call to xnfClose) in order to save changes of the index.
@ref: the xml document (provided by xnfOpen or xnfCreate)
*/
void xnfClose(xmlDocPtr ref);

/*Read operations*/

/* xnfGetDataset
Return a reference to the node containing the requested dataset. return 0 on failure.
Access to arrays is by name (or ID).
The returned xml node must not be released and can be used as the first argument of the following functions.
@ref: a reference to an xml document (provided by xnfOpen or xnfCreate)
@ID: the ID of the given dataset corresponding to the id attribute in the xml index. ID must be an xml name (starting with a letter or underscore or ':',
	containing only letters or digits or '.' '-' '_' ':' )
*/
xmlNodePtr xnfGetDataset(xmlDocPtr ref, const char* ID);

/* xnfGetDimensions
Returns the dimensions of the dataset
@node: a reference returned by xnfGetDataset
@dim: output -> the number of dimensions in the dataset
@dims: int[MaxDim]. output -> dims[0] thru dims[dim-1] contain the dimensions
Result: 0 on success
xnfGetDimensions assumes that length of dims is at least MaxDim.
*/
int xnfGetDimensions(xmlNodePtr node,int* dim,int* dims);

/* xnfArrayGetContents
Returns the dimensions and the content of an array in the dataset node.
@node: a reference returned by xnfGetDataset
@index: 1-based index of the requested array in the dataset
@dim: output -> the number of dimensions in the dataset
@dims: int[MaxDim]. output -> dims[0] thru dims[dim-1] contain the dimensions
@p: output->*p is an array of dims[0]*..*dims[dim-1] numbers. p is allocated with malloc. Use free() to release it
@floatType: the type of the requested numbers
Result: 0 on success
xnfArrayGetContents assumes that length of dims is at least MaxDim.
ex:
	double*p=0;
	int dim,dims[MaxDim];
	err=xnfArrayGetContents( node,1,&dim,dims,(void**)&p,TypeDouble);
*/

int xnfArrayGetContents(xmlNodePtr node,int index,int* dim,int* dims,void**p,NumberType floatType);

/* xnfArrayGetContentsMem
Similar to xnfArrayGetContents, but let you manage the memory.
Fill a block in memory with the content of an array in the dataset node.
The size of the block must be evaluated using xnfGetDimensions.
@node: a reference returned by xnfGetDataset
@index: 1-based index of the requested array in the dataset
@p: an allocated array of dims[0]*..*dims[dim-1] numbers
@floatType: the type of the requested numbers
Result: 0 on success
*/
int xnfArrayGetContentsMem(xmlNodePtr node,int index,void*p,NumberType floatType);

/* xnfArrayGetScale
Returns a scale of the dataset node
@node: a reference returned by xnfGetDataset
@index: a number from 1 to dim
@n: output -> number of element in p: should be dims[index-1]
@p: output -> p has been allocated with malloc. Use free() to release it
@floatType: the type of the requested numbers
Result: 0 on success
*/
int xnfArrayGetScale(xmlNodePtr node,  int index,int* n, void**p,NumberType floatType);

/* xnfArrayGetScaleMem
Similar to xnfArrayGetScale, but let you manage the memory.
@node: a reference returned by xnfGetDataset
@index: a number from 1 to dim
@p: an array of dims[index-1] numbers
@floatType: the type of the requested numbers
Result: 0 on success
*/
int xnfArrayGetScaleMem(xmlNodePtr node,  int index, void*p,NumberType floatType);

/*Write operations*/

/* xnfCreateDataset
Create a new dataset in the xnf index with id 'ID'
@ref: a reference to an xml document (provided by xnfOpen or xnfCreate)
@ID: a string beginning with a letter or underscore (must be an xml name). This ID must be unique in the index file
@dim: the number of dimensions in the dataset
@dims: int[dim]. dims[0] thru dims[dim-1] contain the dimensions of the dataset
return 0 if the id already exists or the index is corrupted, else return the new node which may be used by xnfAddScale or xnfAddArray.
*/
xmlNodePtr xnfCreateDataset(xmlDocPtr ref,const char* ID,int dim,int dims[MaxDim]);

/* xnfAddArray
Add a data node in the dataset cur. cur must be obtained by xnfCreateDataset or xnfGetDataset
@cur: the dataset returned by xnfCreateDataset (or xnfGetDataset)
@size: the number of elements in p, must be dims[0]*..*dims[dim-1]
@p: the pointer containing the array
@floatType: the type of the provided numbers
@href: you may pass 0. href is the name of the file where to add the data. If href is 0, the file name is the id of cur
@data: output, the new data node. Pass 0 if you don't need it
Result: 0 on success
*/

int xnfAddArray(xmlNodePtr cur,int size,void*pp,NumberType floatType,const char*href,xmlNodePtr*data);


/* xnfAddScale
Add a scale for the dimension index.
@cur: the dataset returned by xnfCreateDataset (or xnfGetDataset)
@index: the index of the scale (1-based)
@n: the size of the array
@p: the array providing the scale
@floatType: the type of the provided numbers
Result: 0 on success
*/
int xnfAddScale(xmlNodePtr cur,int index,int n,void*p,NumberType floatType);

/* xnfAddScaleAsRange
Define scales for the dimension index as a pair (start,step).
@cur: the dataset returned by xnfCreateDataset (or xnfGetDataset)
@index: the index of the scale (1-based)
@start: the first value of the array
@step: the increment
Result: 0 on success
*/
int xnfAddScaleAsRange(xmlNodePtr cur,int index,double start , double step);

/* xnfAddScaleAsRef
Define scale for the dimension index as a reference to a 1D array.
@cur: the dataset returned by xnfCreateDataset (or xnfGetDataset)
@index: the index of the scale (1-based)
@ID: the id of existing 1D array
@step: the increment
Result: 0 on success
*/
int xnfAddScaleAsRef(xmlNodePtr cur,int index,const char* ID);

/*Adding references into the index*/
/*
In case where a binary data file already exists, it may be added (or moved) to the 'Contents' directory of the xnf file.
Then the new data may be added with the following function.
*/
/* xnfAddByRef
Create a new data entry in the dataset cur
@cur: the dataset returned by xnfCreateDataset (or xnfGetDataset)
@filename: the filename relative to the 'Contents' directory
@offset: the offset in the file in bytes
@floatType: the type of the numbers in the binary file.
*/
xmlNodePtr xnfAddByRef(xmlNodePtr cur,const char* filename,int offset,NumberType floatType);

/* xnfDatasetRemove
Remove a data or dataset node
@node: a reference to a dataset or an array
 Result: 0 on success
*/
int xnfRemove(xmlNodePtr node);

/* xnfArrayGetRangeContents
 Returns the dimensions and the content of an sub-array in the dataset node.
 @node: a reference returned by xnfGetDataset
 @index: 1-based index of the requested array in the dataset
 @starts: array of the index of the origin of the sub-array (1-based) 
 @lengths: array of the sizes of the requested sub-array
 @dim: output -> the number of dimensions of the resulting array
 @dims: int[MaxDim]. output -> dims[0] thru dims[dim-1] contain the sizes of the sub-array
 @p: output->*p is an array of dims[0]*..*dims[dim-1] numbers. p is allocated with malloc. Use free() to release it
 @floatType: the type of the requested numbers
 Result: 0 on success
 xnfArrayGetRangeContents assumes that length of @starts and lengths is the dimension of the dataset.
 */
int xnfArrayGetRangeContents(xmlNodePtr node,int index,int*starts,int* lengths,int* dim,int* dims,void**p,NumberType floatType);	

/* xnfArrayGetRangeContents
 Similar to xnfArrayGetRangeContents, but let you manage the memory.
 Fill a block in memory with the content of an array in the dataset node.
 The size of the block must be evaluated according to lengths.
 @node: a reference returned by xnfGetDataset
 @index: 1-based index of the requested array in the dataset
 @starts: array of the index of the origin of the sub-array (1-based) 
 @lengths: array of the sizes of the requested sub-array
 @p: an allocated array of lengths[0]*..*lengths[dim-1] numbers
 @floatType: the type of the requested numbers
 Result: 0 on success
 */
int xnfArrayGetRangeContentsMem(xmlNodePtr node,int index,int*starts,int* lengths,void*p,NumberType floatType);
		
enum{
	scale_sizeMissing=100,
	bad_size=101,
	scale_dataMissing=102,
	scale_badstrdata=103,
	scale_startMissing=104,
	scale_steptMissing=105,
	scale_dataEmpty=106,
	invalid_axisIdRef=107,
	cantGetPath=108,/*files*/
	cantOpen=109,
	cantOffset=110,
	cantRead=111,
	cantWrite=112,
	eoferror=113,
	unknownArrayType=114,
	TypeNotFound=115,
	DimNotFound=116,
	BadDim=117,
	invalidNumberOfAxis=118,
	unknownElementType=119,
	endianessMissing=120,
	invalidDims=121,
	invalidStart=122,
	invalidLength=123,
	dataMissing=124,
	IllegalDataIndex=125,
	axisMissing=126,
	IllegalAxisIndex=127,
	invalidType=128,
	incompatibleType=129,
	memFull=130,
	IllegalStartRange=131,
	IllegalEndRange=132
};

#ifdef __cplusplus
}
#endif
#endif
